HEX

( Return bit position of first clear bit and set it )

: ALLOCBIT   ( addr  --  n )
    1  8  0  DO
	OVER  C@  OVER  AND  0=  IF
	    OVER  C@  OR  SWAP  C!  I  LEAVE
	THEN
	2*  LOOP
    DUP  7  >  IF
	2DROP  -1  THEN  ;

( Clear bit at given bit position )

: FREEBIT   ( addr n  --  )
    1  SWAP  LSHIFT  NOT  OVER  C@  AND  SWAP  C!  ;

030C CONSTANT SAREG
030D CONSTANT SXREG
030E CONSTANT SYREG
030F CONSTANT SPREG

( Call machine code subroutine, registers are taken from SAREG, SXREG, SYREG and SPREG and )
( stored back there on return )

CODE SYS   ( addr  --  )
    XSAVE STX,  BOT LDA,  N STA,  BOT 1+ LDA,  N 1+ STA,   ( Store call address )
    HERE  16  +  100  /MOD  # LDA,  PHA,  # LDA,  PHA,   ( Push return address on to stack )
    SXREG LDX,  SYREG LDY,  SPREG  LDA,  PHA,  SAREG LDA,  PLP,
    N ) JMP,   ( RTS returns here because address was pushed above )
    PHP,  SAREG STA,  PLA,  SPREG STA,  SYREG STY,  SXREG STX,  XSAVE LDX,  POP JMP,
END-CODE

( Enter BASIC, perform cold start skipping EXROM checking )

CODE BASIC   (  --  )
    0FF # LDX,  SEI,  TXS,  CLD,  0FD2F JMP,  ( NO RETURN )
END-CODE

( Return jiffy clock as a double )

CODE TIME@   (  --  d )
    DEX,  DEX,  SEI,
    0A2 LDA,  BOT STA,  0A1 LDA,  BOT 1+ STA,  0A0 LDA,  PHA,  TYA,
    CLI,  PUSH JMP,
END-CODE


( Set jiffy clock from double )

CODE TIME!   ( d  --  )
    SEI,
    SEC LDA,  0A2 STA,  SEC 1+ LDA,  0A1 STA,  BOT LDA,  0A0 STA,
    CLI,  POPTWO JMP,
END-CODE


( Low-level device interface )

( Call KERNAL SETNAM function to define file name )
( Prerequisite of $OPEN )

CODE $SETNAM   ( addr count  --  )
    BOT LDA,  PHA,  SEC LDA,  SEC 1+ LDY,  XSAVE STX,  TAX,  PLA,
    0FFBD JSR,  XSAVE LDX,  POPTWO JMP,
END-CODE

( Call KERNAL SETLFS function to define logical file number, )
( device address and secondary address )
( Prerequisite of $OPEN )

CODE $SETLFS   ( sa da lfn  --  )
    BOT LDA,  PHA,  SEC LDA,  SEC 2+ LDY,  XSAVE STX,  TAX,  PLA,
    0FFBA JSR,  XSAVE LDX,  INX,  INX,  POPTWO JMP,

( Return status of last operation, due to a bug in the READST KERNAL )
( function when handling the RS-232 device this is open coded )

LABEL GETST
    0BA LDA,  2 # CMP,  0=   ( Check for RS-232 )
    IF,  0 # LDY,  297 LDA,  297 STY,
    ELSE,  90 LDA,
    THEN,  RTS,
END-CODE

CODE $READST   (  --  st )
    GETST JSR,  PUSH0A JMP,

( Push a combination of the return value of a KERNAL call and the I/O status )
(         15 14-8   7-0 )
(         Cb  .A  STATUS/ )
(                 RSSTAT )
LABEL PUSHSERR
    DEX,  DEX,
LABEL PUTSERR
    CS  IF,  80 # ORA,  ELSE,  0 # LDA,  THEN,  BOT 1+ STA,
    GETST JSR,  BOT STA,  NEXT JMP,
END-CODE

( Call KERNAL OPEN function to open a logical file )
( Requires $SETNAM and $SETLFS to have been called )

CODE $OPEN   (  --  serr )
    XSAVE STX,  0FFC0 JSR,  XSAVE LDX,  PUSHSERR JMP,
END-CODE

( Call KERNAL CLOSE function to close a logical file )

CODE $CLOSE   ( lfn  --  )
    XSAVE STX,  BOT LDA,  0FFC3 JSR,  XSAVE LDX,  POP JMP,
END-CODE

( Call KERNAL CHKOUT function to declare a logical file as output channel )
( All writes - including EMIT - will go to this file until $RSTOUT is )
( called )
( Returns the current output device - usually 3 )
( Prerequisite of $WRITE )

CODE $CHKOUT   ( lfn  --  da serr )
    BOT 1+ STY,  9A LDY,  BOT LDA,  BOT STY,   ( Save current output device )
    XSAVE STX,  TAX,  0FFC9 JSR,  XSAVE LDX,
    PUSHSERR JMP,
END-CODE

( Restore the output channel to a previous device )
( If the current output device is on the serial bus send an UNLISTEN command )

CODE $RSTOUT   ( da  --  )
    3 # LDA,  9A CMP,  CS NOT IF,  0FFAE JSR,  THEN,
    BOT LDA,  9A STA,  POP JMP,
END-CODE

( Call KERNAL CHROUT function to send a character to the output channel )
( Requires $CHKOUT to have been called )

CODE $WRITEB   ( 8b  --  serr )
    BOT LDA,  0FFD2 JSR,  CLC,  PUTSERR JMP,
END-CODE

( Call KERNAL CHROUT function to send a sequence of bytes to the output )
( channel )
( Requires $CHKOUT to have been called )

CODE $WRITE   ( addr count  --  serr )
    2 # LDA,  SETUP JSR,
    BEGIN,  BEGIN,
        N CPY,  0= IF,
        N 1+ LDA,  0= NOT IF,
        SWAP  THEN,
        N 2+ )Y LDA,  0FFD2 JSR,  GETST JSR,
        0= IF,
        INY,
        ROT  0= UNTIL,   ( Repeat until .Y wraps )
        N 3 + INC,  N 1+ DEC,
        ROT  JMP,   ( Repeat )
    SWAP  THEN,  0 # LDA,
    THEN,  PUSH0A JMP,
END-CODE

( Call KERNAL CHKIN function to declare a logical file as input channel )
( All reads - including QUERY - will go to this file until $RSTIN )
( is called )
( Returns the current output device - usually 0 )
( Prerequisite of $READ )

CODE $CHKIN   ( lfn  --  da serr )
    BOT 1+ STY,  99 LDY,  BOT LDA,  BOT STY,   ( Save current input device )
    XSAVE STX,  TAX,  0FFC6 JSR,  XSAVE LDX,
    PUSHSERR JMP,
END-CODE

( Restore the input channel to a previous device )
( If the current input device is on the serial bus send an UNTALK command )

CODE $RSTIN   ( da  --  )
    3 # LDA,  99 CMP,  CS NOT IF,  0FFAB JSR,  THEN,
    BOT LDA,  99 STA,  POP JMP,
END-CODE

( Call KERNAL CHRIN function to receive a sequence of bytes from the input )
( channel )
( Requires $CHKIN to have been called )

CODE $READ   ( addr count1  --  count2 serr )
    2 # LDA,  SETUP JSR,
    DEX,  DEX,  BOT STY,  BOT 1+ STY,   ( zero output count )
    BEGIN,  BEGIN,
        N CPY,  0= IF,
        N 1+ LDA,  0= NOT IF,
        SWAP  THEN,
        0FFCF JSR,  N 2+ )Y STA,
        BOT INC,  0= IF,  BOT 1+ INC,  THEN,
        GETST JSR,
        0= IF,
        INY,
        ROT  0= UNTIL,   ( Repeat until .Y wraps )
    N 3 + INC,  N 1+ DEC,
    ROT  JMP,   ( Repeat )
    SWAP  THEN,  0 # LDA,
    THEN,  PUSH0A JMP,
END-CODE

( Call KERNAL CHRIN function to receive a line of characters from the input )
( channel )
( Requires $CHKIN to have been called )

CODE $READLINE   ( addr count  --  serr )
    2 # LDA,  SETUP JSR,
    BEGIN,  0 # LDA,  INY,
        N CPY,  0= NOT IF,
        CS NOT IF,  SWAP  THEN,
        0FFCF JSR,  0D # CMP,  0= NOT
        IF,  N 2+ )Y STA,  GETST JSR,
        ROT  0= NOT UNTIL,
    ELSE,  DEY,  GETST JSR,
    THEN,
    THEN,
    PHA,  TYA,  0 # LDY,  N 2+ )Y STA,
    PLA,  PUSH0A JMP,
END-CODE

( Call KERNAL CLRCHN function to reset input and output channels )

CODE $CLRCHN   (  --  )
    XSAVE STX,  0FFCC JSR,  XSAVE LDX,  NEXT JMP,
END-CODE

( Call KERNAL SAVE function to write memory region to a device )
( Requires $SETNAM and $SETLFS to have been called )

CODE $SAVE   ( start end  --  serr )
    2 # LDA,  SETUP JSR,  XSAVE STX,
    N LDX,  N 1+ LDY,  N 2+ # LDA,  0FFD8 JSR,
    XSAVE LDX,  PUSHSERR JMP,
END-CODE

( Call KERNAL LOAD function to read memory region from a device )
( Requires $SETNAM and $SETLFS to have been called )

CODE $LOAD   ( start  --  end serr )
    XSAVE STX,  BOT LDA,  BOT 1+ LDY,  TAX,  0 # LDA,  0FFD5 JSR,
    PHA,  TXA,  XSAVE LDX,  BOT STA,  BOT 1+ STY,
    PLA,  PUSHSERR JMP,
END-CODE

( Test KERNAL result & status and QUIT on any error )

: ?IOERR   ( serr  -- )
    0FFBF  AND  ( mask off EOF ) ?DUP  IF
	." ? I/O "  DUP  0<  IF
	    DUP  7F00  AND  8  RSHIFT  .  ELSE
	    23  EMIT  0FF  AND  .  THEN
    QUIT  THEN  ;


( Return a free logical file number )

: NEXT-LFN   (  --  lfn )
    0  BEGIN
	0FF  AND  1+  98  C@  0  ?DO
	    259  I  +  C@  OVER  =  IF  100  OR  LEAVE  THEN
	LOOP
    DUP  100  <  UNTIL  NIP  ;

: LFN2INDEX   ( lfn  --  index )
    100  OR  98  C@  0  ?DO
	259  I  +  C@  OVER  FF  AND  =  IF  DROP  I  LEAVE  THEN
    LOOP
    DUP  100  AND  IF
	DROP  -1
    THEN  ;


8 CONSTANT FIRSTDISK
4 CONSTANT MAXDISKS
VARIABLE DEVICE#
FIRSTDISK  DEVICE#  !
VARIABLE DRIVE#
0  DRIVE#  !

CREATE DISKSTATE
MAXDISKS  2*  ALLOT

: NEXTSA   ( da  --  sa )
    FIRSTDISK  -  DUP
    0  <  0=  IF
	DUP  MAXDISKS  <  IF
	    2*  DISKSTATE  +  ALLOCBIT  2+
	ELSE  DROP  2  THEN
    ELSE  DROP  0  THEN ;

: FREESA   ( sa da  --  )
    FIRSTDISK  -  DUP
    0  <  0=  IF
	DUP  MAXDISKS  <  IF
	    2*  DISKSTATE  +  SWAP  2-  FREEBIT
	ELSE  2DROP  THEN
    ELSE  2DROP  THEN ;


( High-level words resembling ANS Forth File-Access set )

: CLOSE-FILE   ( lfh  --  )
    DUP  LFN2INDEX  DUP  0  <  0=  IF
	DUP  26D  +  C@  1F  AND  SWAP  263  +  C@  FREESA
    ELSE  DROP  THEN
    $CLOSE  ;

: OPEN-FILE   ( addr count sa  --  lfh serr )
    >R  $SETNAM  R>
    DEVICE#  @  DUP  NEXT-LFN  DUP  >R
    $SETLFS
    $OPEN
    R>  SWAP  ;

: READ-FILE   ( addr count1 lfh  --  count2 serr )
    $CHKIN  ?DUP  0=  IF
	>R  $READ  R>  $RSTIN
    ELSE  NIP  0  SWAP
    THEN  ;

: READ-LINE   ( addr count lfh  --  serr )
    $CHKIN  ?DUP  0=  IF
	>R  $READLINE  R>  $RSTIN
    ELSE  NIP
    THEN  ;

: WRITE-FILE   ( addr count lfh  --  serr )
    $CHKOUT  ?DUP  0=  IF
	>R  $WRITE  R>  $RSTOUT
    ELSE  NIP
    THEN  ;

: WRITE-LINE   ( addr count lfn  --  serr )
    $CHKOUT  ?DUP  0=  IF
	>R  $WRITE  ?DUP  0=  IF
	    0D  $WRITEB
	THEN  R>  $RSTOUT
    ELSE  NIP
    THEN  ;

: INCLUDE-FILE   ( lfh  --  )
    $CHKIN  ?IOERR
    >IN  @  >R  INBUF  @  >R   ( Save current source )
    BEGIN
	PAD  0FF  $READLINE
	DUP  0FFBF ( mask off EOF )  AND  0=  IF
	    0  PAD  COUNT  +  C!  0  >IN  !
	    PAD  1+  INBUF  !  >R  INTERPRET  R>
	THEN
	DUP  ?IOERR  UNTIL  $RSTIN
	R>  INBUF  !  R>  >IN  !   ( Restore source )
    ;

: INCLUDED   ( addr count  --  )
    DEVICE#  @  NEXTSA  OPEN-FILE  0=  IF
	DUP  INCLUDE-FILE
	CLOSE-FILE
	THEN  ;

: SAVED   ( auto addr count  --  serr )
    $SETNAM  1  DEVICE#  @  OVER  $SETLFS
    VOC-LINK  @  ,  FENCE  @  ,  ,
#ifdef ROM_TARGET
    [ AUX.ORIGIN ] LITERAL
#else
    0 +ORIGIN
#endif
    HERE $SAVE  -6  DP  +!  ;

: LOADD   ( addr count  --  serr )
    $SETNAM  1  DEVICE#  @  OVER  $SETLFS
    0  $LOAD  DUP  0<  IF
	NIP
    ELSE
	SWAP  6  -  DUP  DP  !  DUP  @  VOC-LINK  !
	2+  DUP  @  FENCE !  2+  @  ?DUP  IF
	    EXECUTE  THEN
	THEN  ;

( Include the file named by the following word )
: INCLUDE
    NAME  COUNT  INCLUDED  ;

( Save the dictionary to the file named by the following word )
: DSAVE
    0  NAME  PAD  OVER  C@  1+  CMOVE>  PAD  COUNT  SAVED  ?IOERR  ;

( Load the dictionary from the file named by the following word )
: DLOAD
    NAME  COUNT  LOADD  ?IOERR  ;

: (RSETCHN)
    $CLRCHN  ;

( {mutable} )
' (RSETCHN) EX-ERROR CATCH
( {immutable} )
